/* -*- Mode: C; tab-width: 4; -*- */
/******************************************************************************
 * Copyright (c) 1996 Netscape Communications. All rights reserved.
 * Copyright (c) 1997 Lawrence T. Hoff (hoff@bnl.gov). All rights reserved.
 ******************************************************************************/
/*
 * UnixShell.c
 *
 * Netscape Client Plugin API
 * - Function that need to be implemented by plugin developers
 *
 * This file defines a "Template" plugin that plugin developers can use
 * as the basis for a real plugin.  This shell just provides empty
 * implementations of all functions that the plugin can implement
 * that will be called by Netscape (the NPP_xxx methods defined in 
 * npapi.h). 
 *
 */

#include <stdio.h>                               /* sprintf(), fprintf(), etc. */
#include <sys/types.h>                           /* pid_t */
#include <unistd.h>                              /* fork(), system(), unlink(), etc. */
#include <signal.h>                              /* SIGHUP */
#include <fcntl.h>                               /* O_RDWR */
#include <sys/wait.h>                            /* wait() */
#include "npapi.h"
#include <sys/time.h>                            /* struct timeval */
#include <sys/resource.h>                        /* maximum # of file descriptors */

/*
** Stuff for the NPP_SetWindow method:
*/
#ifdef XP_UNIX
#include <X11/Xlib.h>                            /* Xlib */
#include <X11/Intrinsic.h>                       /* Xt */
#include <X11/StringDefs.h>                      /* "callback" */
#include <Xm/Form.h>                             /* Motif layout Widget */
#include <Xm/Label.h>                            /* Motif Picture Widget */
#include <Xm/PushB.h>                            /* Motif Button Widget */

#include "xpm.h" /* xpm library functions */
#include "ump.xpm" /* UMP in XPM format */

/* buttons in XPM format */
#include "stopup.xpm"
#include "stopdn.xpm"
#include "playup.xpm"
#include "playdn.xpm"
#include "pauseup.xpm"
#include "pausedn.xpm"
#include "plusup.xpm"
#include "plusdn.xpm"
#include "minusup.xpm"
#include "minusdn.xpm"

#endif /* XP_UNIX */

/* default library, compatible with TIMIDITY */
static const char *timdir = "/usr/local/lib/timidity";

/* resort to slow sampling rate for busy/slow processors */
static int eightKFlag = 0;

/***********************************************************************
 *                                                                     *
 *                                                                     *
 *    Stop playing music, kill the subordinate task                    *
 *                                                                     *
 *                                                                     *
 *                                                                     *
 *                                                                     *
 ***********************************************************************/

static void stopMidiPlayer(pid_t pid){
  if(pid != (pid_t) -1) {
    kill(pid, SIGHUP);
    while(waitpid(pid, 0, WNOHANG)>0); /* clean up after dead child */
  }
}

/***********************************************************************
 *                                                                     *
 *                                                                     *
 *     Function to actually play the music                             *
 *                                                                     *
 *                                                                     *
 *                                                                     *
 *                                                                     *
 ***********************************************************************/

static pid_t startMidiPlayer(const char *filename, int *pipes, int loop, int start){
  
  void timiditySetPipe(int p);
  
  pid_t pid;
  
  
  /* set up IPC */
  (void)pipe(pipes);
  
  /* if we're the parent, return the child ID */
  pid = fork(); if(pid != (pid_t)0) return pid;
  
  /* OK, now let's get ourselves somewhat removed from our parent */
  {
    int fd, sig;
    struct rlimit rlp;
    
    /* start by closing all file descriptors (except stderr, and our pipe) */
    (void) getrlimit(RLIMIT_NOFILE, &rlp);
    for(fd = 3; fd < rlp.rlim_max; fd++) if(fd != pipes[0]) (void) close(fd);
    
    /* now attach stdin and sdtout to /dev/null */
    (void) freopen("/dev/null", "r", stdin);
    (void) freopen("/dev/null", "w", stdout);
    
    /* then remove an netscape-installed signal handlers */
    for(sig=0; sig < 64; sig++)signal(sig, SIG_DFL);
  }
  
  /* Now play the sound file */
  
  {
    static const char *argv[4];
    argv[0] = "timidity";
    argv[1] = "-Od";
    argv[2] = eightKFlag ? "-s8000": "-s22050"; /* check if the user specified 8K rate */
    argv[3] = filename;
    
    /* change to the configuration directory */
    chdir(timdir);
    
    /* tell timidity where to read GUI input */
    timiditySetPipe(pipes[0]);
    
    /* loop forever, until unloaded, wait for play to be enabled */
    do {
      
      signed char c;
      
      /* wait for a new play command before starting */	 
      if(!start) while (read(pipes[0], &c, 1) == 1 && c != ' ');
      
      start = ~0; /* we've started now! */
      
      /* jump to timidity */
      timiditymain(4, argv);
      
      if(!loop){
	/* wait for a new stop command before starting again */
	while (read(pipes[0], &c, 1) == 1 && c != ' ');
	
	/* wait for a new play command before starting again */
	while (read(pipes[0], &c, 1) == 1 && c != ' ');
      }
      
    } while(~0); /* loop forever */
    
  }
  
  /* bye now */
  _exit(0);
  
}

/***********************************************************************
 * Instance state information about the plugin.
 *
 * PLUGIN DEVELOPERS:
 *	Use this struct to hold per-instance information that you'll
 *	need in the various functions in this file.
 ***********************************************************************/

/* we'll read in all the file before processing. Typical MIDI files are smallish,
   so it should be OK */
static const char *filestub = "/tmp/midi";

typedef enum { PLUGIN_PLAYING, PLUGIN_STOPPED, PLUGIN_PAUSED } pluginPlayState;

typedef struct _PluginInstance
{
  int fd; /* file descriptor for temp file */
  const char *filename; /* name of same */
  pid_t pid; /* process ID of sound player (FIXME -- should be global???) */
  int pipes[2]; /* inter-process communication */
  pluginPlayState pluginState; /* stopped, paused, or playing */
  int loop; /* 0 means play only once, non-zero means loop forever */
  
  /* UNIX data members */
#ifdef XP_UNIX
  Widget play, pause, stop, plus, minus; /* button widgets, and their pictures, as well as the logo picture */
  Pixmap logoPixmap, playupPixmap, playdnPixmap, pauseupPixmap, pausednPixmap, stopupPixmap, stopdnPixmap, plusupPixmap, plusdnPixmap, minusupPixmap, minusdnPixmap;
  Display *display; /* needed for XFreePixmap() */
  int width, height; /* how big is the UMP logo? */
#endif /* XP_UNIX */
  
} PluginInstance;


/***********************************************************************
 *
 * Empty implementations of plugin API functions
 *
 * PLUGIN DEVELOPERS:
 *	You will need to implement these functions as required by your
 *	plugin.
 *
 ***********************************************************************/

char*
NPP_GetMIMEDescription(void)
{
  
  static char *desc = 
    "audio/midi:mid,midi:<A HREF=mailto:hoff@bnl.gov>Larry Hoff's</A> "
    "UMP plugin version 1.10;"
    "audio/x-midi:mid,midi:<A HREF=mailto:hoff@bnl.gov>Larry Hoff's</A> "
    "UMP plugin version 1.10;";
  
  return(desc);
}

NPError
NPP_GetValue(void *future, NPPVariable variable, void *value)
{
  
  NPError err = NPERR_NO_ERROR;
  
  static char *desc = 
    
    "This plugins plays MIDI, using the timidity "
    "toolkit, via the standard audio device.<P>"
    "By default, timidity configuration files are in the directory "
    "/usr/local/lib/timidity<P>"
    "Use the TIMID_DIR environment variable to change "
    "the configuration directory.<P>"
    "You may set the environment variable TIMID_8K to "
    "reduce the sampling rate (to 8000 Hz) for busy or slow "
    "processors, or for machines which only support 8KHz.<P>";
  
  
  switch (variable) {
  case NPPVpluginNameString:
    *((char **)value) = "UNIX MIDI plugin";
    break;
  case NPPVpluginDescriptionString:
    *((char **)value) = desc;
    break;
  default:
    err = NPERR_GENERIC_ERROR;
  }
  return err;
}

NPError
NPP_Initialize(void)
{
  
  /* see if the user has specified an alternate directory for timidity */
  if(getenv("TIMID_DIR") != NULL) timdir = getenv("TIMID_DIR");
  
  /* see if the user has specified an alternate directory for timidity */
  if(getenv("TIMID_8K") != NULL) eightKFlag = 1;
  
  return NPERR_NO_ERROR;
}


NPError 
NPP_New(NPMIMEType pluginType,
	NPP instance,
	uint16 mode, /* either NP_EMBED or NP_FULL, we don't care */
	int16 argc,  /* number of name/value pairs */
	char* argn[],
	char* argv[],
	NPSavedData* saved)
{
  PluginInstance* This;
  
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  
  instance->pdata = NPN_MemAlloc(sizeof(PluginInstance));
  
  This = (PluginInstance*) instance->pdata;
  
  if (This == NULL)
    return NPERR_OUT_OF_MEMORY_ERROR;
  
  {
    /* construct a filename unique to this instance to 
       store the data in */
    static char buf[256];
    sprintf(buf, "%s%x.mid", filestub, This);
    This->filename = buf;
    This->pid = -1;
    This->pipes[1] = -1;
    This->pipes[0] = -1;
    This->logoPixmap = -1;
    This->playupPixmap = -1;
    This->playdnPixmap = -1;
    This->pauseupPixmap = -1;
    This->pausednPixmap = -1;
    This->stopupPixmap = -1;
    This->stopdnPixmap = -1;
    This->plusupPixmap = -1;
    This->plusdnPixmap = -1;
    This->minusupPixmap = -1;
    This->minusdnPixmap = -1;
    This->pluginState = PLUGIN_PLAYING;
    This->loop = 0;
  }
  
  {
    /* handle user-overrides */
    
    int i;
    
    
    for(i = 0; i < argc; i++) {
      if(strcasecmp(argn[i], "loop") == 0){
	if(strcasecmp(argv[i], "true") == 0 ||
	   strcasecmp(argv[i], "yes") == 0) {
	  This->loop = ~0;
	}
      }
      else if(strcasecmp(argn[i], "autostart") == 0){
	if(strcasecmp(argv[i], "false") == 0 ||
	   strcasecmp(argv[i], "no") == 0) {
	  This->pluginState = PLUGIN_STOPPED;
	}
	
      }
    }
  }
  
  /* Load associated JAVA class */
  (void) setupLiveConnect(instance, This);
  
  return NPERR_NO_ERROR;
}

NPError 
NPP_Destroy(NPP instance, NPSavedData** save)
{
  PluginInstance* This;
  
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  
  This = (PluginInstance*) instance->pdata;
  
  /* PLUGIN DEVELOPERS:
   *	If desired, call NP_MemAlloc to create a
   *	NPSavedDate structure containing any state information
   *	that you want restored if this plugin instance is later
   *	recreated.
   */
  
  /* quiet things up ASAP */
  if(This->pluginState == PLUGIN_PLAYING){
    
    /* ' ' means toggle pause */
    write(This->pipes[1], " ", 1);
  }
  
  
  if (This != NULL) {
    /* get rid of the temporary file */
    (void) unlink(This->filename);
    
    /* clean up IPC */
    close(This->pipes[1]);
    close(This->pipes[0]);
    
#ifdef XP_UNIX
    if(This->logoPixmap != -1)XFreePixmap(This->display, This->logoPixmap);
    if(This->playupPixmap != -1)XFreePixmap(This->display, This->playupPixmap);
    if(This->playdnPixmap != -1)XFreePixmap(This->display, This->playdnPixmap);
    if(This->pauseupPixmap != -1)XFreePixmap(This->display, This->pauseupPixmap);
    if(This->pausednPixmap != -1)XFreePixmap(This->display, This->pausednPixmap);
    if(This->stopupPixmap != -1)XFreePixmap(This->display, This->stopupPixmap);
    if(This->stopdnPixmap != -1)XFreePixmap(This->display, This->stopdnPixmap);
    if(This->plusupPixmap != -1)XFreePixmap(This->display, This->plusupPixmap);
    if(This->plusdnPixmap != -1)XFreePixmap(This->display, This->plusdnPixmap);
    if(This->minusupPixmap != -1)XFreePixmap(This->display, This->minusupPixmap);
    if(This->minusdnPixmap != -1)XFreePixmap(This->display, This->minusdnPixmap);
#endif
    
    /* stop player */
    stopMidiPlayer(This->pid);
    
    NPN_MemFree(instance->pdata);
    instance->pdata = NULL;
  }
  
  return NPERR_NO_ERROR;
}

/* give buttons the right look for the state of the plugin */

static void setButtons(PluginInstance* This, pluginPlayState state){
  
  switch(This->pluginState = state){
    
  case PLUGIN_PLAYING :
    if(This->playdnPixmap != -1)
      XtVaSetValues(This->play, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->playdnPixmap, NULL);
    if(This->pauseupPixmap != -1)
      XtVaSetValues(This->pause, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->pauseupPixmap, NULL);
    if(This->stopupPixmap != -1)
      XtVaSetValues(This->stop, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->stopupPixmap, NULL);
    break;
    
  case PLUGIN_PAUSED :
    if(This->playupPixmap != -1)
      XtVaSetValues(This->play, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->playdnPixmap, NULL);
    if(This->pausednPixmap != -1)
      XtVaSetValues(This->pause, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->pausednPixmap, NULL);
    if(This->stopupPixmap != -1)
      XtVaSetValues(This->stop, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->stopupPixmap, NULL);
    break;
    
  case PLUGIN_STOPPED :
    if(This->playupPixmap != -1)
      XtVaSetValues(This->play, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->playupPixmap, NULL);
    if(This->pauseupPixmap != -1)
      XtVaSetValues(This->pause, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->pauseupPixmap, NULL);
    if(This->stopdnPixmap != -1)
      XtVaSetValues(This->stop, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->stopdnPixmap, NULL);
    break;
    
  }
  
}

/* handle state changes caused by button pushes */

static void pauseSelected(Widget w, XtPointer client_data, XtPointer call_data){
  
  PluginInstance *This = (PluginInstance *) client_data;
  
  if(This->pluginState != PLUGIN_STOPPED){
    
    /* ' ' means toggle pause */
    write(This->pipes[1], " ", 1);
    
    setButtons(This, This->pluginState == PLUGIN_PAUSED? PLUGIN_PLAYING: PLUGIN_PAUSED);
  }
  
}


static void playSelected(Widget w, XtPointer client_data, XtPointer call_data){
  
  PluginInstance *This = (PluginInstance *) client_data;
  
  if(This->pluginState != PLUGIN_PLAYING){
    
    /* ' ' means toggle pause */
    write(This->pipes[1], " ", 1);
    
    setButtons(This, PLUGIN_PLAYING);
  }
  
}


static void stopSelected(Widget w, XtPointer client_data, XtPointer call_data){
  
  PluginInstance *This = (PluginInstance *) client_data;
  
  if(This->pluginState == PLUGIN_PAUSED){
    
    /* ' q ' means toggle pause, then quit and pause */
    write(This->pipes[1], " q ", 3);
    
    setButtons(This, PLUGIN_STOPPED);
  }
  else if(This->pluginState == PLUGIN_PLAYING){
    
    /* 'q ' means quit and pause */
    write(This->pipes[1], "q ", 2);
    
    setButtons(This, PLUGIN_STOPPED);
  }
  
}


static void plusSelected(Widget w, XtPointer client_data, XtPointer call_data){
  
  PluginInstance *This = (PluginInstance *) client_data;
  
  if(This->plusdnPixmap != -1)
    XtVaSetValues(This->plus, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->plusdnPixmap, NULL);

  /* 'V' means increase volume */
  write(This->pipes[1], "V", 1);

  if(This->plusupPixmap != -1)
    XtVaSetValues(This->plus, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->plusupPixmap, NULL);
  
}


static void minusSelected(Widget w, XtPointer client_data, XtPointer call_data){
  
  PluginInstance *This = (PluginInstance *) client_data;
  
  if(This->minusdnPixmap != -1)
    XtVaSetValues(This->minus, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->minusdnPixmap, NULL);

  /* 'v' means decrease volume */
  write(This->pipes[1], "v", 1);
  
  if(This->minusupPixmap != -1)
    XtVaSetValues(This->minus, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->minusupPixmap, NULL);
  
}


static void writePlayStatus(Widget w, XtPointer closure, XEvent* event, Boolean* b){
  
  NPN_Status((NPP)closure, "Play MIDI music");
  *b = False;
  
}

static void writePauseStatus(Widget w, XtPointer closure, XEvent* event, Boolean* b){
  
  NPN_Status((NPP)closure, "Pause MIDI music");
  *b = False;
  
}

static void writeStopStatus(Widget w, XtPointer closure, XEvent* event, Boolean* b){
  
  NPN_Status((NPP)closure, "Stop MIDI music");
  *b = False;
  
}

static void writePlusStatus(Widget w, XtPointer closure, XEvent* event, Boolean* b){
  
  NPN_Status((NPP)closure, "Increase volume");
  *b = False;
  
}

static void writeMinusStatus(Widget w, XtPointer closure, XEvent* event, Boolean* b){
  
  NPN_Status((NPP)closure, "Decrease volume");
  *b = False;
  
}

static void clearStatus(Widget w, XtPointer closure, XEvent* event, Boolean* b){
  
  NPN_Status((NPP)closure, "");
  *b = False;
  
}


NPError 
NPP_SetWindow(NPP instance, NPWindow* window)
{
  PluginInstance* This;
  
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  
  if (window == NULL)
    return NPERR_NO_ERROR;
  
  This = (PluginInstance*) instance->pdata;
  
  /*
   * PLUGIN DEVELOPERS:
   *	Before setting window to point to the
   *	new window, you may wish to compare the new window
   *	info to the previous window (if any) to note window
   *	size changes, etc.
   */
  
#ifdef XP_UNIX
  {
    
    
    Widget netscape_widget, form, logo;
    XpmAttributes attr;
    NPSetWindowCallbackStruct *winInfo = (NPSetWindowCallbackStruct *)window->ws_info;
    
    This->display = winInfo->display;
    
    /* tell XPM to allocate fairly close (not exact) colors from the window's colormap */
    attr.valuemask = XpmSize|XpmCloseness|XpmAllocCloseColors|XpmExactColors|XpmColormap|XpmDepth|XpmVisual;
    attr.closeness = 30000;
    attr.alloc_close_colors = True;
    attr.exactColors = False;
    attr.colormap = winInfo->colormap;
    attr.depth = winInfo->depth;
    attr.visual = winInfo->visual;
    
    /* while it's at it, tell us what size the logo is */
    /* reuse the pixmap (if possible), assume the colormap and depth will alway be the same */
    /* otherwise we'd have to call XFreePixmap in the XtNdestroyCallback for each widget */
    if(This->playupPixmap == -1) {
      if(XpmCreatePixmapFromData(This->display, (Window) window->window, ump, &This->logoPixmap, NULL, &attr) == XpmSuccess) {
	This->width = attr.width;
	This->height = attr.height;
      }
    }
    
    if(This->playupPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, playup, &This->playupPixmap, NULL, &attr);
    
    if(This->playdnPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, playdn, &This->playdnPixmap, NULL, &attr);
    
    
    if(This->pauseupPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, pauseup, &This->pauseupPixmap, NULL, &attr);
    
    if(This->pausednPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, pausedn, &This->pausednPixmap, NULL, &attr);
    
    
    if(This->stopupPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, stopup, &This->stopupPixmap, NULL, &attr);
    
    if(This->stopdnPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, stopdn, &This->stopdnPixmap, NULL, &attr);
    
    
    if(This->plusupPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, plusup, &This->plusupPixmap, NULL, &attr);
    
    if(This->plusdnPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, plusdn, &This->plusdnPixmap, NULL, &attr);
    
    
    if(This->minusupPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, minusup, &This->minusupPixmap, NULL, &attr);
    
    if(This->minusdnPixmap == -1)
      (void)XpmCreatePixmapFromData(This->display, (Window) window->window, minusdn, &This->minusdnPixmap, NULL, &attr);
    
    /* determine our parentage */
    netscape_widget = XtWindowToWidget(This->display, (Window) window->window);
    
    /* fit our widgets into a form */
    /* no spacing between widgets, we use pixmaps in each widget to define borders */
    form = XtVaCreateWidget("UMP form", xmFormWidgetClass, netscape_widget, 
			    NULL);
    /* Some buttons */
    This->stop = XtVaCreateManagedWidget("stop", xmPushButtonWidgetClass, form, 
					 XmNshadowThickness, 0,
					 XmNhighlightThickness, 0,
					 XmNmarginHeight, 0,
					 XmNmarginWidth, 0,
					 XmNtopAttachment, XmATTACH_FORM,
					 XmNleftAttachment, XmATTACH_FORM,
					 XmNtopOffset, 0,
					 XmNleftOffset, 28,
					 NULL);
    
    This->pause = XtVaCreateManagedWidget("pause", xmPushButtonWidgetClass, form, 
					  XmNshadowThickness, 0,
					  XmNhighlightThickness, 0,
					  XmNmarginHeight, 0,
					  XmNmarginWidth, 0,
					  XmNtopAttachment, XmATTACH_FORM,
					  XmNleftAttachment, XmATTACH_WIDGET,
					  XmNleftWidget, This->stop,
					  XmNtopOffset, 0,
					  XmNleftOffset, 6,
					  NULL);
    
    This->play = XtVaCreateManagedWidget("play", xmPushButtonWidgetClass, form, 
					 XmNshadowThickness, 0,
					 XmNhighlightThickness, 0,
					 XmNmarginHeight, 0,
					 XmNmarginWidth, 0,
					 XmNtopAttachment, XmATTACH_FORM,
					 XmNleftAttachment, XmATTACH_WIDGET,
					 XmNleftWidget, This->pause,
					 XmNtopOffset, 0,
					 XmNleftOffset, 7,
					 NULL);
    
    This->plus = XtVaCreateManagedWidget("plus", xmPushButtonWidgetClass, form, 
					 XmNshadowThickness, 0,
					 XmNhighlightThickness, 0,
					 XmNmarginHeight, 0,
					 XmNmarginWidth, 0,
					 XmNtopAttachment, XmATTACH_FORM,
					 XmNrightAttachment, XmATTACH_FORM,
					 XmNtopOffset, 0,
					 XmNrightOffset, 0,
					 NULL);
    
    This->minus = XtVaCreateManagedWidget("minus", xmPushButtonWidgetClass, form, 
					 XmNshadowThickness, 0,
					 XmNhighlightThickness, 0,
					 XmNmarginHeight, 0,
					 XmNmarginWidth, 0,
					 XmNbottomAttachment, XmATTACH_FORM,
					 XmNrightAttachment, XmATTACH_FORM,
					 XmNbottomOffset, 0,
					 XmNrightOffset, 0,
					 NULL);
    
    /* tell us when the buttons are pushed */
    XtAddCallback(This->play, XmNactivateCallback, playSelected, (XtPointer) This);
    XtAddCallback(This->pause, XmNactivateCallback, pauseSelected, (XtPointer) This);
    XtAddCallback(This->stop, XmNactivateCallback, stopSelected, (XtPointer) This);
    XtAddCallback(This->plus, XmNactivateCallback, plusSelected, (XtPointer) This);
    XtAddCallback(This->minus, XmNactivateCallback, minusSelected, (XtPointer) This);
    
    /* write help text when the mouse is over the button */
    XtAddEventHandler(This->play, EnterWindowMask, False, writePlayStatus, (XtPointer) instance);
    XtAddEventHandler(This->play, LeaveWindowMask, False, clearStatus, (XtPointer) instance);
    
    XtAddEventHandler(This->pause, EnterWindowMask, False, writePauseStatus, (XtPointer) instance);
    XtAddEventHandler(This->pause, LeaveWindowMask, False, clearStatus, (XtPointer) instance);
    
    XtAddEventHandler(This->stop, EnterWindowMask, False, writeStopStatus, (XtPointer) instance);
    XtAddEventHandler(This->stop, LeaveWindowMask, False, clearStatus, (XtPointer) instance);
    
    XtAddEventHandler(This->plus, EnterWindowMask, False, writePlusStatus, (XtPointer) instance);
    XtAddEventHandler(This->plus, LeaveWindowMask, False, clearStatus, (XtPointer) instance);
    
    XtAddEventHandler(This->minus, EnterWindowMask, False, writeMinusStatus, (XtPointer) instance);
    XtAddEventHandler(This->minus, LeaveWindowMask, False, clearStatus, (XtPointer) instance);
    
    
    /* Here's our logo, no border around our nice pixmap */
    logo = XtVaCreateManagedWidget("UMP logo", xmLabelWidgetClass, form, 
				   XmNshadowThickness, 0,
				   XmNhighlightThickness, 0,
				   XmNmarginHeight, 0,
				   XmNmarginWidth, 0,
				   XmNtopAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_FORM,
				   XmNtopOffset, 0,
				   XmNleftOffset, 0,
				   NULL);
    
    /* show the logo */
    if(This->logoPixmap != -1)
      XtVaSetValues(logo, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->logoPixmap, NULL);
    
    /* give buttons the right look and feel */
    setButtons(This, This->pluginState);
    if(This->plusupPixmap != -1)
      XtVaSetValues(This->plus, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->plusupPixmap, NULL);
    if(This->minusupPixmap != -1)
      XtVaSetValues(This->minus, XmNlabelType, XmPIXMAP, XmNlabelPixmap, This->minusupPixmap, NULL);
    
    /* hand over control. Our job is done! */
    XtManageChild(form);
    
    /* jam it to the correct size, otherwise Netscape lets us go wild! */
    XtMakeResizeRequest(form, 
			This->width<window->width?This->width:window->width,
			This->height<window->height?This->height:window->height,
			NULL, NULL);
    
  }
  
#endif /* XP_UNIX */
  
  return NPERR_NO_ERROR;
}


NPError 
NPP_NewStream(NPP instance,
	      NPMIMEType type,
	      NPStream *stream, 
	      NPBool seekable,
	      uint16 *stype)
{
  NPByteRange range;
  PluginInstance* This;
  
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  
  This = (PluginInstance*) instance->pdata;
  
  /* open the file to store data in */
  This->fd = open(This->filename, O_RDWR|O_CREAT, 0666);
  if(This->fd == -1) return NPERR_GENERIC_ERROR;
  
  return NPERR_NO_ERROR;
}


/* PLUGIN DEVELOPERS:
 *	These next 2 functions are directly relevant in a plug-in which
 *	handles the data in a streaming manner. If you want zero bytes
 *	because no buffer space is YET available, return 0. As long as
 *	the stream has not been written to the plugin, Navigator will
 *	continue trying to send bytes.  If the plugin doesn't want them,
 *	just return some large number from NPP_WriteReady(), and
 *	ignore them in NPP_Write().  For a NP_ASFILE stream, they are
 *	still called but can safely be ignored using this strategy.
 */

int32 STREAMBUFSIZE = 0X0FFFFFFF; /* If we are reading from a file in NPAsFile
				   * mode so we can take any size stream in our
				   * write call (since we ignore it) */

int32 
NPP_WriteReady(NPP instance, NPStream *stream)
{
  PluginInstance* This;
  if (instance != NULL)
    This = (PluginInstance*) instance->pdata;
  
  return STREAMBUFSIZE;
}


int32 
NPP_Write(NPP instance, NPStream *stream, int32 offset, int32 len, void *buffer)
{
  if (instance != NULL)
    {
      PluginInstance* This = (PluginInstance*) instance->pdata;
      len = write(This->fd, buffer, len);
      if(len <0) close(This->fd);
    }
  
  return len;		/* The number of bytes accepted */
}


NPError 
NPP_DestroyStream(NPP instance, NPStream *stream, NPError reason)
{
  PluginInstance* This;
  
  if (instance == NULL)
    return NPERR_INVALID_INSTANCE_ERROR;
  This = (PluginInstance*) instance->pdata;
  
  close(This->fd);
  
  /* stop another player in progress ???*/
  stopMidiPlayer(This->pid);
  
  /* yahoo, now we have the data, run timidity */
  This->pid = startMidiPlayer(This->filename, This->pipes, This->loop, This->pluginState == PLUGIN_PLAYING);
  
  NPN_Status(instance, "Playing MIDI file");
  
  return NPERR_NO_ERROR;
}


void 
NPP_StreamAsFile(NPP instance, NPStream *stream, const char* fname)
{
  PluginInstance* This;
  if (instance != NULL)
    This = (PluginInstance*) instance->pdata;
  
  /* just spwan timidity */
  This->pid = startMidiPlayer(fname, This->pipes, This->loop, This->pluginState == PLUGIN_PLAYING);
  
  NPN_Status(instance, "Playing MIDI file");
  
}


void 
NPP_Print(NPP instance, NPPrint* printInfo)
{
  if(printInfo == NULL)
    return;
  
  if (instance != NULL) {
    PluginInstance* This = (PluginInstance*) instance->pdata;
    
    if (printInfo->mode == NP_FULL) {
      /*
       * PLUGIN DEVELOPERS:
       *	If your plugin would like to take over
       *	printing completely when it is in full-screen mode,
       *	set printInfo->pluginPrinted to TRUE and print your
       *	plugin as you see fit.  If your plugin wants Netscape
       *	to handle printing in this case, set
       *	printInfo->pluginPrinted to FALSE (the default) and
       *	do nothing.  If you do want to handle printing
       *	yourself, printOne is true if the print button
       *	(as opposed to the print menu) was clicked.
       *	On the Macintosh, platformPrint is a THPrint; on
       *	Windows, platformPrint is a structure
       *	(defined in npapi.h) containing the printer name, port,
       *	etc.
       */
      
      void* platformPrint =
	printInfo->print.fullPrint.platformPrint;
      NPBool printOne =
	printInfo->print.fullPrint.printOne;
      
      /* Do the default*/
      printInfo->print.fullPrint.pluginPrinted = FALSE;
    }
    else {	/* If not fullscreen, we must be embedded */
      /*
       * PLUGIN DEVELOPERS:
       *	If your plugin is embedded, or is full-screen
       *	but you returned false in pluginPrinted above, NPP_Print
       *	will be called with mode == NP_EMBED.  The NPWindow
       *	in the printInfo gives the location and dimensions of
       *	the embedded plugin on the printed page.  On the
       *	Macintosh, platformPrint is the printer port; on
       *	Windows, platformPrint is the handle to the printing
       *	device context.
       */
      
      NPWindow* printWindow =
	&(printInfo->print.embedPrint.window);
      void* platformPrint =
	printInfo->print.embedPrint.platformPrint;
    }
  }
}

#ifndef INCLUDE_LIVE_CONNECT

setupLiveConnect(NPP instance, PluginInstance *This){}

jref
NPP_GetJavaClass()
{
  return NULL;
}

void
NPP_Shutdown(void)
{
}

#else 

/* IMHO, Live Connect adds lots of overhead and complexity compared to
   the extra functionality it provides (for UMP, anyway). Here we
   allow Java applets or JavaScript to operate the stop/pause/play
   buttons. For this, we must load the "UMP" java class file, which
   must be in the user's CLASSPATH. Then we cannot even operate
   directly on the Widgets, because X gets confused when run from an 
   applet thread, so we use a short (msec) timeout which actually
   does the work. 
   */

#include "UMP.c" /* created by : javac UMP.java; javah -jri -stubs UMP */

setupLiveConnect(NPP instance, PluginInstance *This){
  
  /* store the pointer-to-object for later use (one Java peer per plugin instance) */
  UMP_setThis(NPN_GetJavaEnv(), NPN_GetJavaPeer(instance), (int) This);
  
}

jref
NPP_GetJavaClass()
{
  /* enable Live Connect */
  return use_UMP(NPN_GetJavaEnv());
}

void
NPP_Shutdown(void)
{
  /* all done with Live Connect */
  unuse_UMP(NPN_GetJavaEnv());
}

/* functions for LiveConnect */

static void stopSelectedCallback(XtPointer This, XtIntervalId* id){ stopSelected(NULL, This, NULL); }

void native_UMP_stopSelected(JRIEnv* env, UMP *self, jint This){
  
  PluginInstance *Ump = (PluginInstance *)This;
  
  (void) XtAppAddTimeOut(XtWidgetToApplicationContext(Ump->stop), 1, stopSelectedCallback, (XtPointer) This);
  
}


static void playSelectedCallback(XtPointer This, XtIntervalId* id){ playSelected(NULL, This, NULL); }

/* functions for LiveConnect */
void native_UMP_playSelected(JRIEnv* env, UMP *self, jint This){
  
  PluginInstance *Ump = (PluginInstance *)This;
  
  (void) XtAppAddTimeOut(XtWidgetToApplicationContext(Ump->stop), 1, playSelectedCallback, (XtPointer) This);
  
}


static void pauseSelectedCallback(XtPointer This, XtIntervalId* id){ pauseSelected(NULL, This, NULL); }

/* functions for LiveConnect */
void native_UMP_pauseSelected(JRIEnv* env, UMP *self, jint This){
  
  PluginInstance *Ump = (PluginInstance *)This;
  
  (void) XtAppAddTimeOut(XtWidgetToApplicationContext(Ump->stop), 1, pauseSelectedCallback, (XtPointer) This);
  
}


static void plusSelectedCallback(XtPointer This, XtIntervalId* id){ plusSelected(NULL, This, NULL); }

/* functions for LiveConnect */
void native_UMP_plusSelected(JRIEnv* env, UMP *self, jint This){
  
  PluginInstance *Ump = (PluginInstance *)This;
  
  (void) XtAppAddTimeOut(XtWidgetToApplicationContext(Ump->stop), 1, plusSelectedCallback, (XtPointer) This);
  
}


static void minusSelectedCallback(XtPointer This, XtIntervalId* id){ minusSelected(NULL, This, NULL); }

/* functions for LiveConnect */
void native_UMP_minusSelected(JRIEnv* env, UMP *self, jint This){
  
  PluginInstance *Ump = (PluginInstance *)This;
  
  (void) XtAppAddTimeOut(XtWidgetToApplicationContext(Ump->stop), 1, minusSelectedCallback, (XtPointer) This);
  
}



#endif
